# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause
# See: https://spdx.org/licenses/

from abc import abstractmethod
from dataclasses import dataclass

import numpy as np
import copy
import typing as ty


@dataclass
class EventVolume:
    height: int
    width: int
    polarities: int
    t_start: int = None
    t_stop: int = None


class Transformation:
    """Base class for transformations."""

    @abstractmethod
    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Transform data.

        Parameters
        ----------
        events: np.ndarray
            DVS events

        Returns
        -------
        events: np.ndarray
            Transformed DVS events

        """
        pass

    @abstractmethod
    def determine_output_shape(self, input_shape: EventVolume) -> EventVolume:
        """Transform the shape of the events.

        Parameters
        ----------
        input_shape: EventVolume
            Shape of the incoming events.

        Returns
        -------
        output_shape: EventVolume
            Shape of the outcoming events.
        """
        pass


class Compose:
    def __init__(self, transformations: ty.Iterable[Transformation]):
        """Allows to combine transformations similar to torchvision.Compose.

        Parameters
        ----------
        transformations: Iterable
            Iterable of transformations
        """
        self.transformations = transformations

    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Apply all transformations:

        Parameters
        ----------
        events: np.ndarray
            DVS events

        Returns
        -------
        events: np.ndarray
            Transformed DVS events
        """
        for t in self.transformations:
            events = t(events)
        return events

    def determine_output_shape(self, input_shape: EventVolume):
        """Apply shape transformation of all transfomations in this compose.

        Parameters
        ----------
        input_shape: EventVolume
           Shape of the incoming events.

        Returns
        -------
        output_shape: EventVolume
            Shape of the outcoming events.
        """
        for t in self.transformations:
            input_shape = t.determine_output_shape(input_shape)
        return input_shape


class Downsample(Transformation):
    def __init__(self, factor: ty.Union[float, ty.Dict[str, float]]):
        """Downsamples x and y coordinates by given factor.

        Parameters
        ----------
        factor: float, List[float]
            The factor by which x and y coordinates are downsampled. Can be
            float if the factor shall be the same for x and y, dict defining
            'x' and 'y'.
        input_shape: EventVolume
            Shape of the incoming events.
        output_shape: EventVolume
            If not provided, the transformation calculates the output shape.
            automatically.
        """
        super().__init__()
        if isinstance(factor, float):
            self.factor_x = self.factor_y = factor
        elif isinstance(factor, dict):
            self.factor_x = factor["x"]
            self.factor_y = factor["y"]
        else:
            raise NotImplementedError(
                "factor must be either of type float or dict."
            )

    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Transform data by multiplying height and width with given factor.

        Parameters
        ----------
        events: np.ndarray
            DVS events
        Returns
        -------
        events: np.ndarray
            Transformed DVS events
        """
        events["x"] = (events["x"] * self.factor_x).astype(np.int32)
        events["y"] = (events["y"] * self.factor_y).astype(np.int32)
        return events

    def determine_output_shape(self, input_shape: EventVolume) -> EventVolume:
        """Determine output shape.

         Parameters
         ----------
         input_shape: EventVolume
             Shape of the incoming events.

        Returns
         -------
         output_shape: EventVolume
             Shape of the outcoming events.
        """
        output_shape = copy.deepcopy(input_shape)
        output_shape.width = int(output_shape.width * self.factor_x)
        output_shape.height = int(output_shape.height * self.factor_y)
        return output_shape


class MergePolarities(Transformation):
    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Put all events in one polarity.

        Parameters
        ----------
        events: np.ndarray
            DVS events
        Returns
        -------
        events: np.ndarray
            Transformed DVS events
        """
        events["p"] = (events["p"] * 0).astype(np.int32)
        return events

    def determine_output_shape(self, input_shape: EventVolume) -> EventVolume:
        """Determine output shape.

        Parameters
        ----------
        input_shape: EventVolume
            Shape of the incoming events.

        Returns
        -------
        output_shape: EventVolume
            Shape of the outcoming events.
        """
        output_shape = copy.deepcopy(input_shape)
        output_shape.polarities = 1
        return output_shape


class MirrorHorizontally(Transformation):
    def __init__(self, height: int):
        """Mirror events on a horizontal axis.

        Parameters
        ----------
        height: int
            Height of the original events.
        """
        super().__init__()
        self.height = height

    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Mirror events on a horizontal axis.

        Parameters
        ----------
        events: np.ndarray
            DVS events
        Returns
        -------
        events: np.ndarray
            Transformed DVS events
        """
        events["y"] = self.height - events["y"]
        return events

    def determine_output_shape(self, input_shape: EventVolume) -> EventVolume:
        """Determine output shape.

        Parameters
        ----------
        input_shape: EventVolume
            Shape of the incoming events.

        Returns
        -------
        output_shape: EventVolume
            Shape of the outcoming events.
        """
        return input_shape


class MirrorVertically(Transformation):
    def __init__(self, width: int):
        """Mirror events on a vertical axis.

        Parameters
        ----------
        width: int
           Width of the original events.
        """
        super().__init__()
        self.width = width

    def __call__(self, events: np.ndarray) -> np.ndarray:
        """Mirror events on a vertical axis.

        Parameters
        ----------
        events: np.ndarray
            DVS events
        Returns
        -------
        events: np.ndarray
            Transformed DVS events
        """
        events["x"] = self.width - events["x"]
        return events

    def determine_output_shape(self, input_shape: EventVolume) -> EventVolume:
        """Determine output shape.

        Parameters
        ----------
        input_shape: EventVolume
            Shape of the incoming events.

        Returns
        -------
        output_shape: EventVolume
            Shape of the outcoming events.
        """
        return input_shape



import numpy as np
import time
import cv2

import os
from lava.magma.core.run_configs import Loihi2SimCfg, Loihi2HwCfg
from lava.magma.core.run_conditions import RunSteps, RunContinuous
from lava.magma.core.decorator import implements, requires, tag
from lava.magma.core.model.py.model import PyLoihiProcessModel
from lava.magma.core.model.py.ports import PyOutPort, PyInPort
from lava.magma.core.model.py.type import LavaPyType
from lava.magma.core.process.ports.ports import OutPort, InPort
from lava.magma.core.process.process import AbstractProcess
from lava.magma.core.process.variable import Var
from lava.magma.core.resources import CPU
from lava.magma.core.sync.protocols.loihi_protocol import LoihiProtocol
from lava.proc.io.sink import RingBuffer

from metavision_core.utils import get_sample
from metavision_sdk_cv import TrailFilterAlgorithm

from multiprocessing import Pipe
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import IPython


from metavision_core.event_io import RawReader
from metavision_ml.preprocessing import histo, histo_quantized
from metavision_core.event_io import RawReader

from lava.proc.embedded_io.spike import PyToNxAdapter, NxToPyAdapter
from lava.proc.lif.process import LIF, LogConfig
from lava.proc.dense.process import Dense
from lava.proc.sparse.process import Sparse
import logging

from scipy.sparse import csr_matrix


class EventVisualizer(AbstractProcess):
    """
    Process that receives arbitrary vectors and visualizes them using CV.

    Parameters
    ----------
    shape: tuple, shape of the process
    """

    def __init__(self, shape):
        super().__init__(shape=shape)
        self.shape = shape
        self.s_in = InPort(shape=shape)


@implements(proc=EventVisualizer, protocol=LoihiProtocol)
@requires(CPU)
@tag("floating_pt")
class PyEventVisualizerModel(PyLoihiProcessModel):
    s_in: PyInPort = LavaPyType(PyInPort.VEC_DENSE, np.float32)

    def __init__(self, proc_params):
        super().__init__(proc_params)
        self.shape = proc_params["shape"]
        self.label = "live plot"
        self.max = 0

    def run_spk(self):
        """Receive spikes and store in an internal variable"""
        frame = self.s_in.recv()
        frame.sum(axis=0).sum(axis=0)
        data = np.zeros(self.shape[-2:] + (3,), np.uint8)

        data[:, :, 0] = (
            255 // (frame.max() + 1) * frame[0, 1, :, :].astype(np.uint8)
        )
        data[:, :, 1] = (
            255 // (frame.max() + 1) * frame[0, 0, :, :].astype(np.uint8)
        )
        data[:, :, 2] = (
            255 // (frame.max() + 1) * frame[0, 1, :, :].astype(np.uint8)
        )

        cv2.imshow(self.label, data)
        cv2.waitKey(1)

    def _stop(self):
        cv2.destroyWindow(self.label)
        super()._stop()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause
# See: https://spdx.org/licenses/

import unittest
import numpy as np
import copy

from lava.lib.peripherals.dvs.transformation import (
    Downsample,
    Compose,
    EventVolume,
    MergePolarities,
    MirrorHorizontally,
    MirrorVertically,
)

from metavision_core.event_io import RawReader
from metavision_core.utils import get_sample


SEQUENCE_FILENAME_RAW = "sparklers.raw"
get_sample(SEQUENCE_FILENAME_RAW)


class TestDownsample(unittest.TestCase):
    def test_single_factor(self):
        factor = 0.5

        reader = RawReader(SEQUENCE_FILENAME_RAW)

        delta_t = 100000

        events = reader.load_delta_t(delta_t)
        downsample = Downsample(factor=factor)

        downsampled_events = copy.deepcopy(events)
        downsample(downsampled_events)

        np.testing.assert_equal(
            (events["x"] * factor).astype(int), downsampled_events["x"]
        )
        np.testing.assert_equal(
            (events["y"] * factor).astype(int), downsampled_events["y"]
        )

    def test_per_channel(self):
        factor = {"x": 0.3, "y": 0.7}

        reader = RawReader(SEQUENCE_FILENAME_RAW)

        delta_t = 100000

        events = reader.load_delta_t(delta_t)
        downsample = Downsample(factor=factor)

        downsampled_events = copy.deepcopy(events)
        downsample(downsampled_events)

        np.testing.assert_equal(
            (events["x"] * factor["x"]).astype(int), downsampled_events["x"]
        )
        np.testing.assert_equal(
            (events["y"] * factor["y"]).astype(int), downsampled_events["y"]
        )

    def test_shape_transform(self):
        factor = 0.5

        reader = RawReader(SEQUENCE_FILENAME_RAW)
        height, width = reader.get_size()
        event_volume = EventVolume(height=height, width=width, polarities=2)

        downsample = Downsample(factor=factor)

        output_shape = downsample.determine_output_shape(
            input_shape=event_volume
        )

        self.assertEqual(output_shape.width, int(width) * factor)
        self.assertEqual(output_shape.height, int(height) * factor)


class TestMergePolarities(unittest.TestCase):
    def test_merge(self):
        reader = RawReader(SEQUENCE_FILENAME_RAW)

        delta_t = 100000

        events = reader.load_delta_t(delta_t)
        merge = MergePolarities()

        merged_events = copy.deepcopy(events)
        merge(merged_events)

        self.assertTrue(np.all(merged_events["p"] == 0))

    def test_shape_transform(self):
        reader = RawReader(SEQUENCE_FILENAME_RAW)
        height, width = reader.get_size()
        event_volume = EventVolume(height=height, width=width, polarities=2)

        merge = MergePolarities()
        output_shape = merge.determine_output_shape(input_shape=event_volume)

        self.assertEqual(output_shape.polarities, 1)


class TestMirrorHorizontally(unittest.TestCase):
    def test_mirror(self):
        reader = RawReader(SEQUENCE_FILENAME_RAW)
        height, width = reader.get_size()

        delta_t = 100000

        events = reader.load_delta_t(delta_t)
        transform = MirrorHorizontally(height=height)

        transformed_events = copy.deepcopy(events)
        transform(transformed_events)

        diff = events["y"] - height // 2
        desired_y = height // 2 - diff

        self.assertTrue(np.all(transformed_events["y"] == desired_y))

    def test_shape_transform(self):
        reader = RawReader(SEQUENCE_FILENAME_RAW)
        height, width = reader.get_size()
        event_volume = EventVolume(height=height, width=width, polarities=2)

        transform = MirrorHorizontally(height=height)
        output_shape = transform.determine_output_shape(
            input_shape=event_volume
        )

        self.assertEqual(event_volume, output_shape)


class TestMirrorVertically(unittest.TestCase):
    def test_mirror(self):
        reader = RawReader(SEQUENCE_FILENAME_RAW)
        height, width = reader.get_size()

        delta_t = 100000

        events = reader.load_delta_t(delta_t)
        transform = MirrorVertically(width=width)

        transformed_events = copy.deepcopy(events)
        transform(transformed_events)

        diff = events["x"] - width // 2
        desired_x = width // 2 - diff

        self.assertTrue(np.all(transformed_events["x"] == desired_x))

    def test_shape_transform(self):
        reader = RawReader(SEQUENCE_FILENAME_RAW)
        height, width = reader.get_size()
        event_volume = EventVolume(height=height, width=width, polarities=2)

        transform = MirrorVertically(width=width)
        output_shape = transform.determine_output_shape(
            input_shape=event_volume
        )

        self.assertEqual(event_volume, output_shape)


class TestCompose(unittest.TestCase):
    def test_two_transforms(self):
        downsampling_factor = 0.5

        reader = RawReader(SEQUENCE_FILENAME_RAW)

        delta_t = 100000

        events = reader.load_delta_t(delta_t)
        transforms = Compose(
            [MergePolarities(), Downsample(factor=downsampling_factor)]
        )

        transformed_events = copy.deepcopy(events)
        transforms(transformed_events)

        np.testing.assert_equal(
            (events["x"] * downsampling_factor).astype(int),
            transformed_events["x"],
        )
        np.testing.assert_equal(
            (events["y"] * downsampling_factor).astype(int),
            transformed_events["y"],
        )
        self.assertTrue(np.all(transformed_events["p"] == 0))

    def test_shape_transform(self):
        downsampling_factor = 0.5

        reader = RawReader(SEQUENCE_FILENAME_RAW)
        height, width = reader.get_size()
        event_volume = EventVolume(height=height, width=width, polarities=2)

        transforms = Compose(
            [MergePolarities(), Downsample(factor=downsampling_factor)]
        )

        output_shape = transforms.determine_output_shape(event_volume)

        self.assertEqual(output_shape.width, int(width) * downsampling_factor)
        self.assertEqual(output_shape.height, int(height) * downsampling_factor)
        self.assertEqual(output_shape.polarities, 1)



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause
# See: https://spdx.org/licenses/

import unittest
import os
import time
import numpy as np
import typing as ty

from lava.magma.core.process.process import AbstractProcess
from lava.magma.core.process.ports.ports import InPort
from lava.magma.core.process.variable import Var
from lava.magma.core.sync.protocols.loihi_protocol import LoihiProtocol
from lava.magma.core.decorator import implements, requires
from lava.magma.core.resources import CPU
from lava.magma.core.model.py.model import PyLoihiProcessModel
from lava.magma.core.model.py.type import LavaPyType
from lava.magma.core.model.py.ports import PyInPort
from lava.magma.core.run_configs import Loihi2SimCfg
from lava.magma.core.run_conditions import RunSteps, RunContinuous
from lava.lib.peripherals.dvs.transformation import Compose, Downsample
from lava.lib.peripherals.dvs.prophesee import (
    PropheseeCamera,
    PyPropheseeCameraRawReaderModel,
    PyPropheseeCameraEventsIteratorModel,
    EventsIteratorWrapper)

from metavision_core.utils import get_sample
from metavision_core.event_io import RawReader, EventsIterator
from metavision_sdk_cv import ActivityNoiseFilterAlgorithm


SEQUENCE_FILENAME_RAW = "sparklers.raw"
get_sample(SEQUENCE_FILENAME_RAW)
assert os.path.isfile(SEQUENCE_FILENAME_RAW)

SEQUENCE_FILENAME_DAT = "blinking_leds_td.dat"
get_sample(SEQUENCE_FILENAME_DAT)
assert os.path.isfile(SEQUENCE_FILENAME_DAT)

# Test if camera is connected
try:
    reader = EventsIterator("", delta_t=1)
    del reader
    USE_CAMERA_TESTS = True
except OSError:
    USE_CAMERA_TESTS = False


def get_shape(file_name):
    mv_iterator = EventsIterator(input_path=file_name,
                                 delta_t=1000)
    height, width = mv_iterator.get_size()
    del mv_iterator
    return height, width


class Recv(AbstractProcess):
    """Process that receives arbitrary dense data and stores it in a Var.

    Parameters
    ----------
    shape: tuple
        Shape of the InPort and Var.
    buffer_size: optional, int
        Size of buffer storing received data.
    """

    def __init__(self, shape: ty.Tuple[int], buffer_size: ty.Optional[int] = 1):
        super().__init__(shape=shape, buffer_size=buffer_size)

        self.buffer = Var(shape=(buffer_size,) + shape, init=0)
        self.in_port = InPort(shape=shape)


@implements(proc=Recv, protocol=LoihiProtocol)
@requires(CPU)
class PyRecvProcModel(PyLoihiProcessModel):
    """Receives dense data from PyInPort and stores it in a Var."""

    buffer: np.ndarray = LavaPyType(np.ndarray, np.float32)
    in_port: PyInPort = LavaPyType(PyInPort.VEC_DENSE, np.float32)

    def __init__(self, proc_params: dict) -> None:
        super().__init__(proc_params)
        self._buffer_size = proc_params["buffer_size"]

    def run_spk(self) -> None:
        data = self.in_port.recv()
        self.buffer[(self.time_step - 1) % self._buffer_size] = data


class TestPropheseeCamera(unittest.TestCase):
    def test_init(self):
        """Test that the PropheseeCamera Process is
        instantiated correctly."""
        num_output_time_bins = 3

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW,
            sensor_shape=(height, width),
            num_output_time_bins=num_output_time_bins,
        )

        self.assertIsInstance(camera, PropheseeCamera)

        desired_shape = (num_output_time_bins, 2, height, width)
        self.assertEqual(camera.shape, desired_shape)
        self.assertEqual(camera.s_out.shape, desired_shape)

    def test_invalid_parameters(self):
        """Test that instantiating the PropheseeCamera Process with an invalid
        parameters raises errors."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        n_events = -12
        num_output_time_bins = -1
        biases = {"bias_diff": 80}

        with self.assertRaises(ValueError):
            PropheseeCamera(
                filename=SEQUENCE_FILENAME_RAW,
                sensor_shape=(height, width),
                n_events=n_events,
            )

        with self.assertRaises(ValueError):
            PropheseeCamera(
                filename=SEQUENCE_FILENAME_RAW,
                sensor_shape=(height, width),
                num_output_time_bins=num_output_time_bins,
            )

        with self.assertRaises(ValueError):
            PropheseeCamera(
                filename=SEQUENCE_FILENAME_RAW,
                sensor_shape=(height, width),
                biases=biases,
            )


class TestPyPropheseeCameraModel_EventsIt(unittest.TestCase):
    def test_init(self):
        """Test that the PyPropheseeCameraEventsIteratorModel ProcessModel
        is instantiated correctly."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        transformations = Compose(
            [
                Downsample(factor=0.5),
            ]
        )
        num_output_time_bins = 2

        proc_params = {
            "shape": (num_output_time_bins, 2, height, width),
            "filename": SEQUENCE_FILENAME_RAW,
            "biases": None,
            "filters": [
                ActivityNoiseFilterAlgorithm(
                    width=width, height=height, threshold=1000
                )
            ],
            "mode": "mixed",
            "n_events": 10**8,
            "delta_t": 1000,
            "transformations": transformations
        }

        pm = PyPropheseeCameraEventsIteratorModel(proc_params)

        self.assertIsInstance(pm, PyPropheseeCameraEventsIteratorModel)
        self.assertIsInstance(pm.reader, EventsIteratorWrapper)

    def test_base_functionality_file(self):
        """Test that running a PropheseeCamera works using a data file."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        num_steps = 2

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW, sensor_shape=(height, width)
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = \
            PyPropheseeCameraEventsIteratorModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    def test_base_functionality_dat_file(self):
        """Test that running a PropheseeCamera works
        using a dat data file."""
        # The DAT file should have the same resolution as the RAW file
        height, width = get_shape(SEQUENCE_FILENAME_RAW)
        num_steps = 2

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_DAT,
            sensor_shape=(height, width),
            n_events=1000
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = \
            PyPropheseeCameraEventsIteratorModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    @unittest.skipUnless(USE_CAMERA_TESTS, "Needs live camera")
    def test_base_functionality_camera(self):
        """Test that running a PropheseeCamera works using a camera."""
        num_steps = 2

        camera = PropheseeCamera(filename="", sensor_shape=(720, 1280))

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = \
            PyPropheseeCameraEventsIteratorModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    @unittest.skipUnless(USE_CAMERA_TESTS, "Needs live camera")
    def test_biases(self):
        """Test that setting biases works."""
        num_steps = 2
        biases = {
            "bias_diff": 80,
            "bias_diff_off": 25,
            "bias_diff_on": 140,
            "bias_fo": 74,
            "bias_hpf": 0,
            "bias_refr": 68,
        }

        camera = PropheseeCamera(
            filename="", biases=biases, sensor_shape=(720, 1280)
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = \
            PyPropheseeCameraEventsIteratorModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    def test_filters(self):
        """Test that filters work."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        num_steps = 2
        filters = [
            ActivityNoiseFilterAlgorithm(
                width=width, height=height, threshold=1000
            )
        ]

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW,
            sensor_shape=(height, width),
            filters=filters,
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = \
            PyPropheseeCameraEventsIteratorModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    def test_transformations(self):
        """Test that transformations work."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        num_steps = 2
        transformations = Compose(
            [
                Downsample(factor=0.5),
            ]
        )
        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW,
            sensor_shape=(height, width),
            transformations=transformations,
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = \
            PyPropheseeCameraEventsIteratorModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    def test_data_received(self):
        """Test that the PropheseeCamera process correctly reads and sends
        data."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)
        num_steps = 1

        transformations = Compose(
            [
                Downsample(factor=0.1),
            ]
        )

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW,
            sensor_shape=(height, width),
            transformations=transformations,
        )

        recv = Recv(shape=camera.s_out.shape, buffer_size=num_steps)

        camera.s_out.connect(recv.in_port)

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = \
            PyPropheseeCameraEventsIteratorModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        recv_data = recv.buffer.get()
        camera.stop()

        self.assertTrue(np.any(recv_data > 0))

    def test_pause(self):
        """Test pausing the network does not cause any harm.
        Data will get droped for the pause duration."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW, sensor_shape=(height, width)
        )

        run_condition = RunContinuous()
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = \
            PyPropheseeCameraEventsIteratorModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        time.sleep(0.1)
        camera.pause()
        time.sleep(0.1)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        time.sleep(0.1)
        camera.stop()


class TestPyPropheseeCameraModel_RawReader(unittest.TestCase):
    def test_init(self):
        """Test that the PyPropheseeCameraRawReaderModel ProcessModel is instantiated
        correctly."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        transformations = Compose(
            [
                Downsample(factor=0.5),
            ]
        )
        num_output_time_bins = 2

        proc_params = {
            "shape": (num_output_time_bins, 2, height, width),
            "filename": SEQUENCE_FILENAME_RAW,
            "biases": None,
            "filters": [
                ActivityNoiseFilterAlgorithm(
                    width=width, height=height, threshold=1000
                )
            ],
            "n_events": 10**8,
            "transformations": transformations,
            "sync_time": True,
            "flatten": False
        }

        pm = PyPropheseeCameraRawReaderModel(proc_params)

        self.assertIsInstance(pm, PyPropheseeCameraRawReaderModel)
        self.assertIsInstance(pm.reader, RawReader)

    def test_base_functionality_file(self):
        """Test that running a PropheseeCamera works using a data file."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        num_steps = 2

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW, sensor_shape=(height, width)
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = PyPropheseeCameraRawReaderModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    def test_base_functionality_dat_file(self):
        """Test that running a PropheseeCamera works
        using a dat data file."""
        # The DAT file should have the same resolution as the RAW file

        height, width = get_shape(SEQUENCE_FILENAME_DAT)
        num_steps = 2

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_DAT, sensor_shape=(height, width)
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = PyPropheseeCameraRawReaderModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    @unittest.skipUnless(USE_CAMERA_TESTS, "Needs live camera")
    def test_base_functionality_camera(self):
        """Test that running a PropheseeCamera works using a camera."""
        num_steps = 2

        camera = PropheseeCamera(filename="", sensor_shape=(720, 1280))

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = PyPropheseeCameraRawReaderModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    @unittest.skipUnless(USE_CAMERA_TESTS, "Needs live camera")
    def test_biases(self):
        """Test that setting biases works."""

        num_steps = 2
        biases = {
            "bias_diff": 80,
            "bias_diff_off": 25,
            "bias_diff_on": 140,
            "bias_fo": 74,
            "bias_hpf": 0,
            "bias_refr": 68,
        }

        camera = PropheseeCamera(
            filename="", biases=biases, sensor_shape=(720, 1280)
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = PyPropheseeCameraRawReaderModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    def test_filters(self):
        """Test that filters work."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        num_steps = 2
        filters = [
            ActivityNoiseFilterAlgorithm(
                width=width, height=height, threshold=1000
            )
        ]

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW,
            sensor_shape=(height, width),
            filters=filters,
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = PyPropheseeCameraRawReaderModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    def test_transformations(self):
        """Test that transformations work."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        num_steps = 2
        transformations = Compose(
            [
                Downsample(factor=0.5),
            ]
        )
        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW,
            sensor_shape=(height, width),
            transformations=transformations,
        )

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = PyPropheseeCameraRawReaderModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        camera.stop()

    def test_data_received(self):
        """Test that the PropheseeCamera process correctly reads and sends
        data."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)
        num_steps = 1

        transformations = Compose(
            [
                Downsample(factor=0.1),
            ]
        )

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW,
            sensor_shape=(height, width),
            transformations=transformations,
        )

        recv = Recv(shape=camera.s_out.shape, buffer_size=num_steps)

        camera.s_out.connect(recv.in_port)

        run_condition = RunSteps(num_steps=num_steps)
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = PyPropheseeCameraRawReaderModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        recv_data = recv.buffer.get()
        camera.stop()

        self.assertTrue(np.any(recv_data > 0))

    def test_pause(self):
        """Test pausing the network does not cause any harm.
        Data will get droped for the pause duration."""

        height, width = get_shape(SEQUENCE_FILENAME_RAW)

        camera = PropheseeCamera(
            filename=SEQUENCE_FILENAME_RAW, sensor_shape=(height, width)
        )

        run_condition = RunContinuous()
        custom_proc_model_map = {}
        custom_proc_model_map[PropheseeCamera] = PyPropheseeCameraRawReaderModel
        run_cfg = Loihi2SimCfg(exception_proc_model_map=custom_proc_model_map)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        time.sleep(0.1)
        camera.pause()
        time.sleep(0.1)
        camera.run(condition=run_condition, run_cfg=run_cfg)
        time.sleep(0.1)
        camera.stop()


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause
# See: https://spdx.org/licenses/

import unittest
import sys
from pathlib import Path

from lava.lib.peripherals.realsense.realsense import RealSense
from lava.magma.core.run_configs import Loihi2SimCfg
from lava.magma.core.run_conditions import RunSteps

USE_CAMERA_TESTS = False
try:
    import pyrealsense2 as rs

    for device in rs.context().devices:
        color_sensor = device.first_color_sensor()
        depth_sensor = device.first_depth_sensor()

        if color_sensor is not None and depth_sensor is not None:
            USE_CAMERA_TESTS = True
            break
except ImportError:
    print("Need `pyrealsense2` library installed.", file=sys.stderr)
    exit(1)


class TestRealSense(unittest.TestCase):
    @unittest.skipUnless(USE_CAMERA_TESTS,
                         "Requires a Realsense camera to be connected.")
    def test_init_camera(self) -> None:
        """Test that the Realsense Process is instantiated correctly when
        reading from camera."""
        realsense = RealSense()

        if realsense.bgr_2d_shape != realsense.depth_2d_shape:
            self.assertNotEqual(realsense.bgr_out_port.shape[:2],
                                realsense.depth_out_port.shape)

        realsense = RealSense(align_depth_to_bgr=True)

        self.assertEqual(realsense.bgr_out_port.shape[:2],
                         realsense.depth_out_port.shape)

    def test_init_files(self) -> None:
        """Test that the Realsense Process is instantiated correctly when
        reading from files."""
        directory_path = Path(__file__).resolve().parent / "recording"

        realsense = RealSense(directory_path=str(directory_path),
                              png_prefix="bgr_",
                              exr_prefix="depth_")

        self.assertEqual(realsense.bgr_out_port.shape, (480, 640, 3))
        self.assertEqual(realsense.depth_out_port.shape, (240, 320))

    def test_init_files_invalid_parameters(self) -> None:
        """Test that initializing the Realsense Process reading from
        files, with invalid parameters, raises errors."""
        with self.assertRaises(NotADirectoryError):
            _ = RealSense(directory_path="invalid")

        directory_path = Path(__file__).resolve().parent / "recording"

        with self.assertRaises(FileNotFoundError):
            _ = RealSense(directory_path=str(directory_path),
                          png_prefix="invalid",
                          exr_prefix="depth_")

        with self.assertRaises(FileNotFoundError):
            _ = RealSense(directory_path=str(directory_path),
                          png_prefix="bgr_",
                          exr_prefix="invalid")


class TestLoihiDensePyRealSensePM(unittest.TestCase):
    @unittest.skipUnless(USE_CAMERA_TESTS,
                         "Requires a Realsense camera to be connected.")
    def test_run_camera(self) -> None:
        """Test that the Realsense Process runs correctly when reading from
        camera."""
        realsense = RealSense()
        realsense.run(condition=RunSteps(num_steps=2), run_cfg=Loihi2SimCfg())
        realsense.stop()

    def test_run_files(self) -> None:
        """Test that the Realsense Process runs correctly when reading from
        files."""
        directory_path = Path(__file__).resolve().parent / "recording"

        realsense = RealSense(directory_path=str(directory_path),
                              png_prefix="bgr_",
                              exr_prefix="depth_")
        realsense.run(condition=RunSteps(num_steps=2), run_cfg=Loihi2SimCfg())
        realsense.stop()


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause
# See: https://spdx.org/licenses/

import sys

try:
    from metavision_core.event_io import (EventsIterator, RawReader,
                                          EventDatReader)
    from metavision_ml.preprocessing.event_to_tensor import histo_quantized
except ImportError:
    print("Need `metavision` library installed.", file=sys.stderr)
    exit(1)

import numpy as np
import time
from threading import Thread

from lava.magma.core.decorator import implements, requires, tag
from lava.magma.core.model.py.model import PyLoihiProcessModel
from lava.magma.core.model.py.ports import PyOutPort
from lava.magma.core.model.py.type import LavaPyType
from lava.magma.core.process.ports.ports import OutPort
from lava.magma.core.process.process import AbstractProcess
from lava.magma.core.resources import CPU
from lava.magma.core.sync.protocols.loihi_protocol import LoihiProtocol

from lava.lib.peripherals.dvs.transformation import Compose, EventVolume


class PropheseeCamera(AbstractProcess):
    """
    Process that receives events from Prophesee device and sends them out as a
    histogram.

    Parameters
    ----------
    filename: str
        String to filename if reading from a RAW/DAT file or empty string for
        using a camera.
    biases: dict
        Dictionary of biases for the DVS Camera.
    filters: list
        List containing metavision filters.
    mode: str
        String to specify to load events by numbers of events, timeslice or
        the first met criterion. (Choice of "mixed", "delta_t", or "n_events").
        Default is "mixed".
    n_events: int
        Number of events in the timeslice.
    delta_t: int
        Duration of served event slice in us.
    transformations: Compose
        Transformations to be applied to the events before sending them out.
    num_output_time_bins: int
        The number of output time bins to use for the ToFrame transformation.
    sync_time: bool
        Whether to synchronize camera timestamps with Loihi 2 runtime.
    flatten: bool
        Whether to send flattened output.
    """

    def __init__(
            self,
            sensor_shape: tuple,
            filename: str = "",
            biases: dict = None,
            filters: list = [],
            mode: str = "mixed",
            n_events: int = 10 ** 8,
            delta_t: int = 1000,
            transformations: Compose = None,
            num_output_time_bins: int = 1,
            out_shape: tuple = None,
            sync_time: bool = True,
            flatten: bool = False
    ):
        if not isinstance(n_events, int) or n_events < 0:
            raise ValueError(
                "n_events must be a positive integer value."
            )

        if (
                not isinstance(num_output_time_bins, int)
                or num_output_time_bins < 0
        ):
            raise ValueError(
                "num_output_time_bins must be a positive integer value."
            )

        if biases is not None and not filename == "":
            raise ValueError("Cant set biases if reading from file.")

        self.filename = filename
        self.biases = biases
        self.sync_time = sync_time
        self.flatten = flatten
        self.mode = mode
        self.n_events = n_events
        self.delta_t = delta_t
        self.filters = filters
        self.transformations = transformations
        self.num_output_time_bins = num_output_time_bins

        height, width = sensor_shape

        if out_shape is not None:
            self.shape = out_shape
        # Automatically determine out_shape
        else:
            event_shape = EventVolume(height=height, width=width, polarities=2)
            if transformations is not None:
                event_shape = self.transformations.determine_output_shape(
                    event_shape
                )
            self.shape = (
                num_output_time_bins,
                event_shape.polarities,
                event_shape.height,
                event_shape.width,
            )

        # Check whether provided transformation is valid
        if self.transformations is not None:
            try:
                # Generate some artificial data
                n_random_spikes = 1000
                test_data = np.zeros(
                    n_random_spikes,
                    dtype=np.dtype(
                        [("y", int), ("x", int), ("p", int), ("t", int)]
                    ),
                )
                test_data["x"] = np.random.rand(n_random_spikes) * width
                test_data["y"] = np.random.rand(n_random_spikes) * height
                test_data["p"] = np.random.rand(n_random_spikes) * 2
                test_data["t"] = np.sort(np.random.rand(n_random_spikes) * 1e6)

                # Transform data
                self.transformations(test_data)
                if len(test_data) > 0:
                    volume = np.zeros(self.shape, dtype=np.uint8)
                    histo_quantized(test_data, volume, np.max(test_data["t"]))

            except Exception:
                raise Exception(
                    "Your transformation is not compatible with the provided \
                    data."
                )
        num_neurons = np.prod(self.shape)
        if self.flatten:
            self.s_out = OutPort(shape=(num_neurons,))
        else:
            self.s_out = OutPort(shape=self.shape)

        super().__init__(
            shape=self.shape,
            biases=self.biases,
            filename=self.filename,
            filters=self.filters,
            mode=self.mode,
            n_events=self.n_events,
            delta_t=self.delta_t,
            transformations=self.transformations,
            num_output_time_bins=self.num_output_time_bins,
            sync_time=self.sync_time,
            flatten=self.flatten
        )


class EventsIteratorWrapper():
    """
    PropheseeEventsIterator class for PropheseeCamera which will create a
    thread in the background to always grab events within a time window and
    put them in a buffer.

    delta_t: Control the size of the time window for collecting events,
    and generate frames for the events in this time window.Increase delta_t,
    a frame can capture more events, but when you move fast, there will be
    too many events, at this time, you should consider reducing delta_t.
    If delta_t is too large and greater than time interval between two steps
    of processmodel, the thread will recv the old events data from the
    previous time step.

    Parameters
    ----------
    device: str
        String to filename if reading from a RAW/DAT file or empty string for
        using a camera.
    sensor_shape: (int, int)
        Shape of the camera sensor or file recording.
    mode: str
        String to specify to load events by numbers of events, timeslice or
        the first met criterion. (Choice of "mixed", "delta_t", or "n_events").
        Default is "mixed".
    n_events: int
        Number of events in the timeslice.
    delta_t: int
        Duration of served event slice in us.
    biases: list
        Bias settings for camera.
    """

    def __init__(self,
                 device: str,
                 sensor_shape: tuple,
                 mode: str = "mixed",
                 n_events: int = 1000,
                 delta_t: int = 1000,
                 biases: dict = None, ):
        self.true_height, self.true_width = sensor_shape

        self.mv_iterator = EventsIterator(input_path=device, mode=mode,
                                          n_events=n_events, delta_t=delta_t)

        if biases is not None:
            # Setting Biases for DVS camera
            device_biases = self.mv_iterator.reader.device.get_i_ll_biases()
            for k, v in biases.items():
                device_biases.set(k, v)

        self.thread = Thread(target=self.recv_from_dvs)
        self.stop = False
        self.res = np.zeros(((self.true_width, self.true_height) + (1,) + (1,)),
                            dtype=np.dtype([("y", int), ("x", int),
                                            ("p", int), ("t", int)]))

    def start(self):
        self.thread.start()

    def join(self):
        self.stop = True
        self.thread.join()

    def get_events(self):
        return self.res

    def recv_from_dvs(self):
        for evs in self.mv_iterator:
            self.res = evs

            if self.stop:
                return


@implements(proc=PropheseeCamera, protocol=LoihiProtocol)
@requires(CPU)
@tag("floating_pt", "event_it")
class PyPropheseeCameraEventsIteratorModel(PyLoihiProcessModel):
    s_out: PyOutPort = LavaPyType(PyOutPort.VEC_DENSE, np.int32)

    def __init__(self, proc_params):
        super().__init__(proc_params)
        self.shape = proc_params["shape"]
        (
            self.num_output_time_bins,
            self.polarities,
            self.height,
            self.width,
        ) = self.shape
        self.filename = proc_params["filename"]
        self.filters = proc_params['filters']
        self.mode = proc_params['mode']
        self.n_events = proc_params['n_events']
        self.delta_t = proc_params['delta_t']
        self.biases = proc_params['biases']
        self.transformations = proc_params['transformations']
        self.sensor_shape = (self.height,
                             self.width)

        self.reader = EventsIteratorWrapper(
            device=self.filename,
            sensor_shape=self.sensor_shape,
            mode=self.mode,
            n_events=self.n_events,
            delta_t=self.delta_t,
            biases=self.biases)
        self.reader.start()

        self.volume = np.zeros(
            (
                self.num_output_time_bins,
                self.polarities,
                self.height,
                self.width,
            ),
            dtype=np.uint8,
        )

    def run_spk(self):
        """
        Load events from DVS, apply filters and transformations and send
        spikes as frame
        """
        events = self.reader.get_events()

        # Apply filters to events
        for filter in self.filters:
            events_out = filter.get_empty_output_buffer()
            filter.process_events(events, events_out)
            events = events_out

        if len(self.filters) > 0:
            events = events.numpy()

        # Transform events
        if self.transformations is not None and len(events) > 0:
            self.transformations(events)

        # Transform to frame
        if len(events) > 0:
            histo_quantized(events, self.volume, 1000, reset=True)
            frames = self.volume
        else:
            frames = np.zeros(self.s_out.shape)

        self.s_out.send(frames)

    def _pause(self):
        """Pause was called by the runtime"""
        super()._pause()
        self.t_pause = time.time_ns()

    def _stop(self):
        """
        Stop was called by the runtime.
        Helper thread for DVS is also stopped.
        """
        self.reader.join()
        super()._stop()


@implements(proc=PropheseeCamera, protocol=LoihiProtocol)
@requires(CPU)
@tag("floating_pt", "raw_reader")
class PyPropheseeCameraRawReaderModel(PyLoihiProcessModel):
    s_out: PyOutPort = LavaPyType(PyOutPort.VEC_DENSE, np.int32)

    def __init__(self, proc_params):
        super().__init__(proc_params)
        self.shape = proc_params["shape"]
        (
            self.num_output_time_bins,
            self.polarities,
            self.height,
            self.width,
        ) = self.shape
        self.filename = proc_params["filename"]
        self.filters = proc_params["filters"]
        self.n_events = proc_params["n_events"]
        self.biases = proc_params["biases"]
        self.transformations = proc_params["transformations"]
        self.flatten = proc_params["flatten"]
        self.sync_time = proc_params["sync_time"]

        if self.filename.split('.')[-1] == 'dat':
            self.reader = EventDatReader(self.filename)
        else:
            self.reader = RawReader(self.filename,
                                    max_events=self.n_events)

        if self.biases is not None:
            # Setting Biases for DVS camera
            device_biases = self.reader.device.get_i_ll_biases()
            for k, v in self.biases.items():
                device_biases.set(k, v)

        self.volume = np.zeros(
            (
                self.num_output_time_bins,
                self.polarities,
                self.height,
                self.width,
            ),
            dtype=np.uint8,
        )
        self.t_pause = None
        self.t_last_iteration = None

    def run_spk(self):
        """Load events from DVS, apply filters and transformations and send
        spikes as frame"""

        if self.sync_time:
            if self.t_pause is None:
                self.t_pause = time.time_ns()

            if self.t_last_iteration is None:
                self.t_last_iteration = time.time_ns()

            # Time passed since last iteration
            t_now = time.time_ns()

            # Load new events since last iteration
            if self.t_pause > self.t_last_iteration:
                # Runtime was paused in the meantime
                delta_t = np.max(
                    [10000, (self.t_pause - self.t_last_iteration) // 1000]
                )
                delta_t_drop = np.max([10000, (t_now - self.t_pause) // 1000])

                events = self.reader.load_delta_t(delta_t)
                _ = self.reader.load_delta_t(delta_t_drop)
            else:
                # Runtime was not paused in the meantime
                delta_t = np.max([10000,
                                  (t_now - self.t_last_iteration) // 1000])
                events = self.reader.load_delta_t(delta_t)
        else:
            delta_t = 10000
            events = self.reader.load_delta_t(delta_t)

        # Apply filters to events
        for filter in self.filters:
            events_out = filter.get_empty_output_buffer()
            filter.process_events(events, events_out)
            events = events_out

        if len(self.filters) > 0:
            events = events.numpy()

        # Transform events
        if self.transformations is not None and len(events) > 0:
            self.transformations(events)

        # Transform to frame
        if len(events) > 0:
            histo_quantized(events, self.volume, delta_t, reset=True)
            frames = self.volume
        else:
            frames = np.zeros(self.s_out.shape)

        # Send
        if self.flatten:
            frames = frames.flatten()
        self.s_out.send(frames)
        self.t_last_iteration = time.time_ns()

    def _pause(self):
        """Pause was called by the runtime"""
        super()._pause()
        self.t_pause = time.time_ns()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause
# See: https://spdx.org/licenses/

import numpy as np
import typing as ty
from pathlib import Path
import sys
import os

from lava.magma.core.process.process import AbstractProcess
from lava.magma.core.process.ports.ports import OutPort
from lava.magma.core.sync.protocols.loihi_protocol import LoihiProtocol
from lava.magma.core.model.py.ports import PyOutPort
from lava.magma.core.model.py.type import LavaPyType
from lava.magma.core.resources import CPU
from lava.magma.core.decorator import implements, requires
from lava.magma.core.model.py.model import PyLoihiProcessModel

try:
    import pyrealsense2 as rs
except ImportError:
    print("Need `pyrealsense2` library installed.", file=sys.stderr)
    exit(1)


os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
import cv2  # noqa: E402


class RealSense(AbstractProcess):
    """Process that reads BGR+depth frames, either from a RealSense camera
    directly, or from recorded PNG+EXR files.

    If directory_path is not given, the Process will read and send BGR and
    depth frames from a RealSense camera.

    If directory_path is given, the Process will read and send BGR and depth
    frames from PNG and EXR files contained in the directory which the path
    points to.
    In this case, png_prefix and exr_prefix have to be provided to inform the
    Process on how the files are named.

    Parameters
    ----------
    align_depth_to_bgr: bool, optional
        Boolean flag controlling whether or not to align depth frames to BGR
        frames.
    directory_path: str, optional
        Path to directory that contains the images PNG+EXR files.
        If not given, will use the camera instead.
    png_prefix: str, optional
        Prefix of PNG files (recording of BGR frames).
    exr_prefix: str, optional
        Prefix of EXR files (recording of depth frames).
    """

    def __init__(self,
                 align_depth_to_bgr: ty.Optional[bool] = False,
                 directory_path: ty.Optional[str] = None,
                 png_prefix: ty.Optional[str] = "",
                 exr_prefix: ty.Optional[str] = "") -> None:
        super().__init__(
            align_depth_to_bgr=align_depth_to_bgr,
            directory_path=directory_path,
            png_prefix=png_prefix,
            exr_prefix=exr_prefix
        )

        if directory_path is None:
            pipeline = rs.pipeline()
            config = rs.config()
            config.enable_stream(rs.stream.depth)
            config.enable_stream(rs.stream.color)

            profile = pipeline.start(config)
            device = profile.get_device()
            bgr_sensor = device.first_color_sensor()
            depth_sensor = device.first_depth_sensor()

            if bgr_sensor is None:
                raise ValueError("No BGR sensor was found.")
            if depth_sensor is None:
                raise ValueError("No depth sensor was found.")

            bgr_stream = \
                profile.get_stream(rs.stream.color).as_video_stream_profile()
            depth_stream = \
                profile.get_stream(rs.stream.depth).as_video_stream_profile()

            self.bgr_2d_shape = (bgr_stream.height(), bgr_stream.width())
            self.bgr_3d_shape = self.bgr_2d_shape + (3,)
            self.depth_2d_shape = (depth_stream.height(), depth_stream.width())

            pipeline.stop()
        else:
            dir_path = Path(directory_path)

            if not dir_path.is_dir():
                raise NotADirectoryError(f"{dir_path} is not a directory.")

            bgr_sample_path = Path(directory_path) / f"{png_prefix}1.png"
            depth_sample_path = Path(directory_path) / f"{exr_prefix}1.exr"

            if not bgr_sample_path.is_file():
                raise FileNotFoundError(f"{bgr_sample_path} is not a file.")
            if not depth_sample_path.is_file():
                raise FileNotFoundError(f"{depth_sample_path} is not a file.")

            bgr_sample_frame = cv2.imread(str(bgr_sample_path))
            depth_sample_frame = \
                cv2.imread(str(depth_sample_path),
                           cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH)

            self.bgr_3d_shape = bgr_sample_frame.shape
            self.bgr_2d_shape = self.bgr_3d_shape[:2]
            self.depth_2d_shape = depth_sample_frame.shape

            if align_depth_to_bgr:
                raise NotImplementedError("Aligning Depth frames to BGR "
                                          "frames is not implemented.")

        bgr_out_shape = self.bgr_3d_shape
        depth_out_shape = self.depth_2d_shape

        if align_depth_to_bgr:
            depth_out_shape = self.bgr_2d_shape

        self.proc_params["bgr_2d_shape"] = self.bgr_2d_shape
        self.proc_params["bgr_3d_shape"] = self.bgr_3d_shape
        self.proc_params["depth_2d_shape"] = self.depth_2d_shape

        self.bgr_out_port = OutPort(shape=bgr_out_shape)
        self.depth_out_port = OutPort(shape=depth_out_shape)


@implements(proc=RealSense, protocol=LoihiProtocol)
@requires(CPU)
class LoihiDensePyRealSensePM(PyLoihiProcessModel):
    bgr_out_port: PyOutPort = LavaPyType(PyOutPort.VEC_DENSE, int)
    depth_out_port: PyOutPort = LavaPyType(PyOutPort.VEC_DENSE, int)

    def __init__(self, proc_params: dict) -> None:
        super().__init__(proc_params)

        self._bgr_2d_shape = proc_params["bgr_2d_shape"]
        self._bgr_3d_shape = proc_params["bgr_3d_shape"]
        self._depth_2d_shape = proc_params["depth_2d_shape"]
        self._align_depth_to_bgr = proc_params["align_depth_to_bgr"]
        self._directory_path = proc_params["directory_path"]
        self._png_prefix = proc_params["png_prefix"]
        self._exr_prefix = proc_params["exr_prefix"]

        if self._directory_path is None:
            self._pipeline = rs.pipeline()

            config = rs.config()
            config.enable_stream(rs.stream.depth)
            config.enable_stream(rs.stream.color)

            pipeline_wrapper = rs.pipeline_wrapper(self._pipeline)
            pipeline_profile = config.resolve(pipeline_wrapper)
            device = pipeline_profile.get_device()
            device.hardware_reset()

            self._pipeline.start(config)
            self._align = rs.align(rs.stream.color)

    def run_spk(self) -> None:
        bgr_image, depth_image = self._get_frames()

        self.bgr_out_port.send(bgr_image)
        self.depth_out_port.send(depth_image)

    def _get_frames(self) -> ty.Tuple[np.ndarray, np.ndarray]:
        """Get BGR and depth frames.

        If directory_path is None, get frames from camera.
        If directory_path is not None, get frames from files
        contained in the directory which the path points to.

        Returns
        ----------
        bgr_frame : np.ndarray
            BGR frame.
        depth_frame : np.ndarray
            Depth frame.
        """
        if self._directory_path is None:
            return self._get_frames_from_camera()
        else:
            return self._get_frames_from_directory()

    def _get_frames_from_camera(self) -> ty.Tuple[np.ndarray, np.ndarray]:
        """Get BGR and depth frames from camera.

        Returns
        ----------
        bgr_frame : np.ndarray
            BGR frame.
        depth_frame : np.ndarray
            Depth frame.
        """
        frameset = self._pipeline.wait_for_frames()
        if self._align_depth_to_bgr:
            frameset = self._align.process(frameset)

        bgr_frame = np.array(frameset.get_color_frame().get_data())
        depth_frame = np.array(frameset.get_depth_frame().get_data())

        return bgr_frame, depth_frame

    def _get_frames_from_directory(self) -> ty.Tuple[np.ndarray, np.ndarray]:
        """Get BGR and depth frames from files.

        Returns
        ----------
        bgr_frame : np.ndarray
            BGR frame.
        depth_frame : np.ndarray
            Depth frame.
        """
        bgr_frame_path = Path(self._directory_path) / f"{self._png_prefix}" \
                                                      f"{self.time_step}.png"
        depth_frame_path = Path(self._directory_path) / f"{self._exr_prefix}" \
                                                        f"{self.time_step}.exr"

        bgr_frame = cv2.imread(str(bgr_frame_path))
        depth_frame = cv2.imread(str(depth_frame_path),
                                 cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH)

        return bgr_frame, depth_frame

    def _stop(self) -> None:
        if self._directory_path is None:
            self._pipeline.stop()

        super()._stop()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause
# See: https://spdx.org/licenses/

import cv2
import numpy as np

from lava.magma.core.process.process import AbstractProcess
from lava.magma.core.process.ports.ports import InPort
from lava.magma.core.sync.protocols.loihi_protocol import LoihiProtocol
from lava.magma.core.model.py.ports import PyInPort
from lava.magma.core.model.py.type import LavaPyType
from lava.magma.core.resources import CPU
from lava.magma.core.decorator import implements, requires
from lava.magma.core.model.py.model import PyLoihiProcessModel


class OpenCVDisplay(AbstractProcess):
    """ Process that display incoming image with OpenCV.

    This process continuously display grayscale/RGB images coming through its
    InPort. The number of input ports is decided by the number of process models
    connected to this opencv process model.

    Parameters
    ----------
    shape: tuple
        Shape of the InPort.
        (height, width) or (height, width, 1) for grayscale images.
        (height, width, 3) for RGB images.
    opencv_window_name: str
        Name of OpenCV window.
    """
    def __init__(self,
                 shape: tuple,
                 opencv_window_name: str = "Image Display"):
        super().__init__(shape=shape,
                         opencv_window_name=opencv_window_name)
        if len(shape) not in (2, 3):
            raise ValueError("Shape must be in the format (height, width) for \
                             grayscale or (height, width, 3) for RGB.")
        if len(shape) == 3 and shape[2] not in (1, 3):
            raise ValueError("For a 3-channel image, shape must be \
                             (height, width, 3). For grayscale, use \
                             (height, width) or (height, width, 1).")
        self.frame_port = InPort(shape=shape)


@implements(proc=OpenCVDisplay, protocol=LoihiProtocol)
@requires(CPU)
class OpenCVDisplayPM(PyLoihiProcessModel):
    frame_port: PyInPort = LavaPyType(PyInPort.VEC_DENSE, float)

    def __init__(self, proc_params):
        super().__init__(proc_params)
        self._shape = proc_params['shape']
        self._window_name = proc_params['opencv_window_name']
        cv2.namedWindow(self._window_name)

    def display(self, image):
        """ Visualize images on OpenCV windows.

        Takes a NumPy image array formatted as RGBA and sends to OpenCV for
        visualization. RGBA image will be converted to grayscale by summing
        the color channels. If it's a 2D image, no changes will be made.

        Parameters
        ----------
        image: np.ndarray
            Image to display.
        """
        image = np.array(np.sum(image, axis=2),
                         dtype=np.uint8)
        cv2.imshow(self._window_name, image)
        cv2.waitKey(1)

    def run_spk(self):
        frame = self.frame_port.recv()
        self.display(frame)

    def _stop(self):
        cv2.destroyAllWindow()
        super()._stop()
